﻿using System;
using System.Linq;
using System.ServiceModel.Description;
using System.Xml.Linq;
using Microsoft.Crm.Sdk.Samples;
using Microsoft.Crm.Services.Utility;
using Microsoft.Xrm.Sdk.Client;
using Microsoft.Xrm.Sdk.Discovery;

// http://dotnetfollower.com
namespace dotNetFollower
{
    public class ServerConnectionEx : ServerConnection
    {
        /// <summary>
        /// Parses xml string and builds a Configuration object with connection parameters
        /// </summary>
        public Configuration Parse(string xml)
        {
            // read xml and fetch out parameters
            XElement configurationsFromFile = XElement.Parse(xml);
            var xmlConfig = configurationsFromFile.Nodes().First() as XElement;

            var newConfig = new Configuration();
            
            var serverAddress = xmlConfig.Element("ServerAddress");
            if (serverAddress != null && !String.IsNullOrEmpty(serverAddress.Value))
                newConfig.ServerAddress = serverAddress.Value;

            var organizationName = xmlConfig.Element("OrganizationName");
            if (organizationName != null && !String.IsNullOrEmpty(organizationName.Value))
                newConfig.OrganizationName = organizationName.Value;

            var discoveryUri = xmlConfig.Element("DiscoveryUri");
            if (discoveryUri != null && !String.IsNullOrEmpty(discoveryUri.Value))
                newConfig.DiscoveryUri = new Uri(discoveryUri.Value);

            var organizationUri = xmlConfig.Element("OrganizationUri");
            if (organizationUri != null && !String.IsNullOrEmpty(organizationUri.Value))
                newConfig.OrganizationUri = new Uri(organizationUri.Value);

            var homeRealmUri = xmlConfig.Element("HomeRealmUri");
            if (homeRealmUri != null && !String.IsNullOrEmpty(homeRealmUri.Value))
                newConfig.HomeRealmUri = new Uri(homeRealmUri.Value);

            var vendpointType = xmlConfig.Element("EndpointType");
            if (vendpointType != null)
                newConfig.EndpointType = RetrieveAuthenticationType(vendpointType.Value);

            var xElement = xmlConfig.Element("Credentials");
            if (xElement != null && xElement.HasElements)
                newConfig.Credentials = ParseInCredentials(xmlConfig.Element("Credentials"), newConfig.EndpointType);

            if (newConfig.EndpointType == AuthenticationProviderType.LiveId)
                newConfig.DeviceCredentials = DeviceIdManager.LoadOrRegisterDevice();

            var userPrincipalName = xmlConfig.Element("UserPrincipalName");
            if (userPrincipalName != null && !String.IsNullOrWhiteSpace(userPrincipalName.Value))
                newConfig.UserPrincipalName = userPrincipalName.Value;

            // save new configuration in the config property as the property 
            // is used by original ServerConnection
            config = newConfig;

            // get OrganizationUri in case it is not set
            if (newConfig.OrganizationUri == null)
            {
                if(string.IsNullOrEmpty(newConfig.OrganizationName))
                    throw new ApplicationException("At least one of OrganizationUri and OrganizationName must be spesified in the xml configuration.");
                GetOrganizationInfo();
            }

            return newConfig;
        }

        public static void DoInOrganizationServiceProxyContext(Configuration configuration, Action<OrganizationServiceProxyEx> action)
        {
            using (var proxy = GetOrganizationProxy(configuration))
            {
                proxy.EnableProxyTypes();
                action(proxy);
            }
        }

        public void DoInOrganizationServiceProxyContext(string configXml, Action<OrganizationServiceProxyEx> action)
        {
            Configuration configuration = Parse(configXml);
            DoInOrganizationServiceProxyContext(configuration, action);
        }

        /// <summary>
        /// Gets Organization Web Service url by friend or unique name of the organization
        /// </summary>
        protected void GetOrganizationInfo()
        {
            using (DiscoveryServiceProxy serviceProxy = GetDiscoveryProxy())
            {
                // Obtain organization information from the Discovery service. 
                if (serviceProxy != null)
                {
                    // Obtain information about the organizations that the system user belongs to.
                    OrganizationDetailCollection orgs = DiscoverOrganizations(serviceProxy);

                    if(orgs.Count == 0)
                        throw new ApplicationException("User does not belong to any organizations on the specified server");

                    // look for a specified organization
                    foreach (var organizationDetail in orgs)
                    {
                        if (string.Equals(organizationDetail.UniqueName, config.OrganizationName, StringComparison.OrdinalIgnoreCase) ||
                            string.Equals(organizationDetail.FriendlyName, config.OrganizationName, StringComparison.OrdinalIgnoreCase))
                        {
                            config.OrganizationUri  = new Uri(organizationDetail.Endpoints[EndpointType.OrganizationService]);
                            config.OrganizationName = organizationDetail.FriendlyName;
                            return;
                        }
                    }

                    throw new ApplicationException("Couldn't find the " + config.OrganizationName + "organization");
                }
                
                throw new ApplicationException("An invalid server name was specified.");
            }
        }

        /// <summary>
        /// Parses xml element to get user name and password
        /// </summary>
        protected static ClientCredentials ParseInCredentials(XElement credentials, AuthenticationProviderType endpointType)
        {
            ClientCredentials result = null;

            if (credentials.HasElements)
            {
                result = new ClientCredentials();

                switch (endpointType)
                {
                    case AuthenticationProviderType.ActiveDirectory:
                        {
                            result.Windows.ClientCredential = new System.Net.NetworkCredential()
                                {
                                    UserName = credentials.Element("UserName").Value,
                                    Domain   = credentials.Element("Domain").Value,
                                    Password = credentials.Element("Password").Value
                                };
                            break;
                        }
                    case AuthenticationProviderType.LiveId:
                    case AuthenticationProviderType.Federation:
                    case AuthenticationProviderType.OnlineFederation:
                        {
                            result.UserName.UserName = credentials.Element("UserName").Value;
                            result.UserName.Password = credentials.Element("Password").Value;
                            break;
                        }
                }
            }

            return result;
        }
    }
}
